package org.apache.bcel.classfile;

import java.util.Objects;
import org.apache.bcel.generic.Type;
import java.io.IOException;
import java.io.DataInput;
import org.apache.bcel.util.BCELComparator;

public final class Method extends FieldOrMethod
{
    public static final Method[] EMPTY_ARRAY;
    private static BCELComparator bcelComparator;
    static final Method[] EMPTY_METHOD_ARRAY;
    private ParameterAnnotationEntry[] parameterAnnotationEntries;
    
    public static BCELComparator getComparator() {
        return Method.bcelComparator;
    }
    
    public static void setComparator(final BCELComparator comparator) {
        Method.bcelComparator = comparator;
    }
    
    public Method() {
    }
    
    Method(final DataInput file, final ConstantPool constantPool) throws IOException, ClassFormatException {
        super(file, constantPool);
    }
    
    public Method(final int accessFlags, final int nameIndex, final int signatureIndex, final Attribute[] attributes, final ConstantPool constantPool) {
        super(accessFlags, nameIndex, signatureIndex, attributes, constantPool);
    }
    
    public Method(final Method c) {
        super(c);
    }
    
    @Override
    public void accept(final Visitor v) {
        v.visitMethod(this);
    }
    
    public Method copy(final ConstantPool constantPool) {
        return (Method)this.copy_(constantPool);
    }
    
    @Override
    public boolean equals(final Object obj) {
        return Method.bcelComparator.equals(this, obj);
    }
    
    public Type[] getArgumentTypes() {
        return Type.getArgumentTypes(this.getSignature());
    }
    
    public Code getCode() {
        for (final Attribute attribute : super.getAttributes()) {
            if (attribute instanceof Code) {
                return (Code)attribute;
            }
        }
        return null;
    }
    
    public ExceptionTable getExceptionTable() {
        for (final Attribute attribute : super.getAttributes()) {
            if (attribute instanceof ExceptionTable) {
                return (ExceptionTable)attribute;
            }
        }
        return null;
    }
    
    public LineNumberTable getLineNumberTable() {
        final Code code = this.getCode();
        if (code == null) {
            return null;
        }
        return code.getLineNumberTable();
    }
    
    public LocalVariableTable getLocalVariableTable() {
        final Code code = this.getCode();
        if (code == null) {
            return null;
        }
        return code.getLocalVariableTable();
    }
    
    public ParameterAnnotationEntry[] getParameterAnnotationEntries() {
        if (this.parameterAnnotationEntries == null) {
            this.parameterAnnotationEntries = ParameterAnnotationEntry.createParameterAnnotationEntries(this.getAttributes());
        }
        return this.parameterAnnotationEntries;
    }
    
    public Type getReturnType() {
        return Type.getReturnType(this.getSignature());
    }
    
    @Override
    public int hashCode() {
        return Method.bcelComparator.hashCode(this);
    }
    
    @Override
    public String toString() {
        final String access = Utility.accessToString(super.getAccessFlags());
        ConstantUtf8 c = super.getConstantPool().getConstantUtf8(super.getSignatureIndex());
        String signature = c.getBytes();
        c = super.getConstantPool().getConstantUtf8(super.getNameIndex());
        final String name = c.getBytes();
        signature = Utility.methodSignatureToString(signature, name, access, true, this.getLocalVariableTable());
        final StringBuilder buf = new StringBuilder(signature);
        for (final Attribute attribute : super.getAttributes()) {
            if (!(attribute instanceof Code) && !(attribute instanceof ExceptionTable)) {
                buf.append(" [").append(attribute).append("]");
            }
        }
        final ExceptionTable e = this.getExceptionTable();
        if (e != null) {
            final String str = e.toString();
            if (!str.isEmpty()) {
                buf.append("\n\t\tthrows ").append(str);
            }
        }
        return buf.toString();
    }
    
    static {
        EMPTY_ARRAY = new Method[0];
        Method.bcelComparator = new BCELComparator() {
            @Override
            public boolean equals(final Object o1, final Object o2) {
                final Method THIS = (Method)o1;
                final Method THAT = (Method)o2;
                return Objects.equals(THIS.getName(), THAT.getName()) && Objects.equals(THIS.getSignature(), THAT.getSignature());
            }
            
            @Override
            public int hashCode(final Object o) {
                final Method THIS = (Method)o;
                return THIS.getSignature().hashCode() ^ THIS.getName().hashCode();
            }
        };
        EMPTY_METHOD_ARRAY = new Method[0];
    }
}
